<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8" />
    <title>Password Reset</title>
  </head>
  <body>
    <p>Hello.</p>
    <p>
      You have requested a password reset. Please click on the link below to
      reset your password:
    </p>
    <p>
      <a id="reset-url" href="{{ reset_url }}">Reset Password</a
      >
    </p>
    <p>If you did not make this request, please contact your administrator.</p>
  </body>
</html>
